==> tinyerp-3.0.tar.gz: 3.0/client/bin/widget/form/textbox_tag.py <==
##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#					Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

#
# TODO: recode the whole pango code
# 

import gtk
from gtk import glade
import pango

import wid_int
import common

class textbox_tag(wid_int.wid_int):
	desc_to_attr_table = {
		'family':[pango.AttrFamily,""],
		'style':[pango.AttrStyle,pango.STYLE_NORMAL],
		'variant':[pango.AttrVariant,pango.VARIANT_NORMAL],
		'weight':[pango.AttrWeight,pango.WEIGHT_NORMAL],
		'stretch':[pango.AttrStretch,pango.STRETCH_NORMAL],
		}
	pango_translation_properties={
			pango.ATTR_SIZE : 'size',
			pango.ATTR_WEIGHT: 'weight',
			pango.ATTR_UNDERLINE: 'underline',
			pango.ATTR_STRETCH: 'stretch',
			pango.ATTR_VARIANT: 'variant',
			pango.ATTR_STYLE: 'style',
			pango.ATTR_SCALE: 'scale',
			pango.ATTR_STRIKETHROUGH: 'strikethrough',
			pango.ATTR_RISE: 'rise',
			}
	attval_to_markup={
			'underline':{pango.UNDERLINE_SINGLE:'single',
						 pango.UNDERLINE_DOUBLE:'double',
						 pango.UNDERLINE_LOW:'low',
						 pango.UNDERLINE_NONE:'none'},
			'stretch':{pango.STRETCH_ULTRA_EXPANDED:'ultraexpanded',
					   pango.STRETCH_EXPANDED:'expanded',
					   pango.STRETCH_EXTRA_EXPANDED:'extraexpanded',
					   pango.STRETCH_EXTRA_CONDENSED:'extracondensed',
					   pango.STRETCH_ULTRA_CONDENSED:'ultracondensed',
					   pango.STRETCH_CONDENSED:'condensed',
					   pango.STRETCH_NORMAL:'normal',
					   },
			'variant':{pango.VARIANT_NORMAL:'normal',
					   pango.VARIANT_SMALL_CAPS:'smallcaps',
					   },
			'style':{pango.STYLE_NORMAL:'normal',
					 pango.STYLE_OBLIQUE:'oblique',
					 pango.STYLE_ITALIC:'italic',
					 },
			'stikethrough':{1:'true',
							True:'true',
							0:'false',
							False:'false'},
			}

	def __init__(self, parent, attrs={}):
		wid_int.wid_int.__init__(self, parent, attrs)
		self.win_gl = glade.XML(common.terp_path("terp.glade"),"widget_textbox_tag")
		self.widget = self.win_gl.get_widget('widget_textbox_tag')
		self.tv = self.win_gl.get_widget('widget_textbox_tag_tv')
		self.widget.show_all()
		self.tags = {}
		self.buf = self.tv.get_buffer()
		self.tagdict = {}


		#a,t,s = pango.parse_markup('<b>bold</b>',u'0')

		#tags=self.get_tags_from_attrs(font,None,attr)
		#self.tag_widgets[tuple(tags)]=widg

		#self.win_gl.signal_connect('on_toggle_bold_toggled', self._toggle)
		#self.win_gl.signal_connect('on_toggle_italic_toggled', self._toggle)
		#self.win_gl.signal_connect('on_toggle_underline_toggled', self._toggle)



	def get_tags_from_attrs (self, font,lang,attrs):
		tags = []
		if font:
			font,fontattrs = self.fontdesc_to_attrs(font)
			fontdesc = font.to_string()
			if fontattrs:
				attrs.extend(fontattrs)
			if fontdesc and fontdesc!='Normal':
				if not self.tags.has_key(font.to_string()):
					tag=self.buf.create_tag()
					tag.set_property('font-desc',font)
					if not self.tagdict.has_key(tag): self.tagdict[tag]={}
					self.tagdict[tag]['font_desc']=font.to_string()
					self.tags[font.to_string()]=tag
				tags.append(self.tags[font.to_string()])
		if lang:
			if not self.tags.has_key(lang):
				tag = self.buf.create_tag()
				tag.set_property('language',lang)
				self.tags[lang]=tag
			tags.append(self.tags[lang])
		if attrs:
			for a in attrs:
				if a.type == pango.ATTR_FOREGROUND:
					gdkcolor = self.pango_color_to_gdk(a.color)
					key = 'foreground%s'%self.color_to_hex(gdkcolor)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property('foreground-gdk',gdkcolor)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]]['foreground']="#%s"%self.color_to_hex(gdkcolor)
					tags.append(self.tags[key])
				if a.type == pango.ATTR_BACKGROUND:
					gdkcolor = self.pango_color_to_gdk(a.color)
					tag.set_property('background-gdk',gdkcolor)
					key = 'background%s'%self.color_to_hex(gdkcolor)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property('background-gdk',gdkcolor)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]]['background']="#%s"%self.color_to_hex(gdkcolor)
					tags.append(self.tags[key])
				if self.pango_translation_properties.has_key(a.type):
					prop=self.pango_translation_properties[a.type]
					#print 'setting property %s of %s (type: %s)'%(prop,a,a.type)
					val=getattr(a,'value')
					#tag.set_property(prop,val)
					mval = val
					if self.attval_to_markup.has_key(prop):
						#print 'converting ',prop,' in ',val
						if self.attval_to_markup[prop].has_key(val):
							mval = self.attval_to_markup[prop][val]
					key="%s%s"%(prop,val)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property(prop,val)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]][prop]=mval
					tags.append(self.tags[key])
				else:
					print "Don't know what to do with attr %s"%a
		return tags

	def get_tags (self):
		tagdict = {}
		for pos in range(self.buf.get_char_count()):
			iter=self.buf.get_iter_at_offset(pos)
			for tag in iter.get_tags():
				if tagdict.has_key(tag):
					if tagdict[tag][-1][1] == pos - 1:
						tagdict[tag][-1] = (tagdict[tag][-1][0],pos)
					else:
						tagdict[tag].append((pos,pos))
				else:
					tagdict[tag]=[(pos,pos)]
		return tagdict

	def get_text (self):
		tagdict=self.get_tags()
		txt = self.buf.get_text(self.buf.get_start_iter(),self.buf.get_end_iter())
		cuts = {}
		for k,v in tagdict.items():
			stag,etag = self.tag_to_markup(k)
			for st,end in v:
				if cuts.has_key(st): cuts[st].append(stag) #add start tags second
				else: cuts[st]=[stag]
				if cuts.has_key(end+1): cuts[end+1]=[etag]+cuts[end+1] #add end tags first
				else: cuts[end+1]=[etag]
		last_pos = 0
		outbuff = ""
		cut_indices = cuts.keys()
		cut_indices.sort()
		for c in cut_indices:
			if not last_pos==c:
				outbuff += txt[last_pos:c]
				last_pos = c
			for tag in cuts[c]:
				outbuff += tag
		outbuff += txt[last_pos:]
		return outbuff

	def tag_to_markup (self, tag):
		stag = "<span"
		for k,v in self.tagdict[tag].items():
			stag += ' %s="%s"'%(k,v)
		stag += ">"
		return stag,"</span>"

	def fontdesc_to_attrs (self,font):
		nicks = font.get_set_fields().value_nicks
		attrs = []
		for n in nicks:
			if self.desc_to_attr_table.has_key(n):
				Attr,norm = self.desc_to_attr_table[n]
				# create an attribute with our current value
				attrs.append(Attr(getattr(font,'get_%s'%n)()))
				# unset our font's value
				getattr(font,'set_%s'%n)(norm)
		return font,attrs

	def pango_color_to_gdk (self, pc):
		return gtk.gdk.Color(pc.red,pc.green,pc.blue)

	def color_to_hex (self, color):
		hexstring = ""
		for col in 'red','green','blue':
			hexfrag = hex(getattr(color,col)/(16*16)).split("x")[1]
			if len(hexfrag)<2: hexfrag = "0" + hexfrag
			hexstring += hexfrag
		return hexstring

	def apply_font_and_attrs (self, font, attrs):
		tags = self.get_tags_from_attrs(font,None,attrs)
		for t in tags: self.apply_tag(t)

	def remove_font_and_attrs (self, font, attrs):
		tags = self.get_tags_from_attrs(font,None,attrs)
		for t in tags: self.remove_tag(t)

	def setup_default_tags (self):
		self.italics = self.get_tags_from_attrs(None,None,[pango.AttrStyle('italic')])[0]
		self.bold = self.get_tags_from_attrs(None,None,[pango.AttrWeight('bold')])[0]
		self.underline = self.get_tags_from_attrs(None,None,[pango.AttrUnderline('single')])[0]

	def get_selection (self):
		bounds = self.buf.get_selection_bounds()
		if not bounds:
			iter=self.buf.get_iter_at_mark(self.insert)
			if iter.inside_word():
				start_pos = iter.get_offset()
				iter.forward_word_end()
				word_end = iter.get_offset()
				iter.backward_word_start()
				word_start = iter.get_offset()
				iter.set_offset(start_pos)
				bounds = (self.buf.get_iter_at_offset(word_start),
						  self.buf.get_iter_at_offset(word_end))
		return bounds

	def apply_tag (self, tag):
		selection = self.get_selection()
		if selection:
			self.buf.apply_tag(tag,*selection)

	def remove_tag (self, tag):
		selection = self.get_selection()
		if selection:
			self.buf.remove_tag(tag,*selection)

	def remove_all_tags (self):
		selection = self.get_selection()
		if selection:
			for t in self.tags.values():
				self.buf.remove_tag(t,*selection)

	def set_text (self, txt):
		buf = self.tv.get_buffer()
		parsed,txt,separator = pango.parse_markup(txt,u'0')
		attrIter = parsed.get_iterator()
		while True:
			range=attrIter.range()
			font,lang,attrs = attrIter.get_font()
			tags = self.get_tags_from_attrs(font,lang,attrs)
			text = txt[range[0]:range[1]]
			if tags:
				buf.insert_with_tags(buf.get_end_iter(),text,*tags)
			else:
				buf.insert_with_tags(buf.get_end_iter(),text)
			if not attrIter.next():
				break

	#
	# Button callback
	#
	def setup_widget_from_pango (self, widg, markupstring):
		"""setup widget from a pango markup string"""
		#font = pango.FontDescription(fontstring)
		a,t,s = pango.parse_markup(markupstring,u'0')
		ai=a.get_iterator()
		font,lang,attrs=ai.get_font()
		self.setup_widget(widg,font,attrs)

	def setup_widget (self, widg, font, attr):
		tags=self.get_tags_from_attrs(font,None,attr)
		self.tag_widgets[tuple(tags)]=widg
		widg.connect('toggled',self._toggle,tags)

	def _toggle (self, widget, tags):
		#print '_toggle called'
		if self.internal_toggle: return
		#print 'not internal'
		if widget.get_active():
			#print 'applying'
			for t in tags: self.apply_tag(t)
		else:
			#print 'removing'
			for t in tags: self.remove_tag(t)

	def _mark_set_cb (self, buffer, iter, mark, *params):
		#print '_mark_set_cb'
		if mark==self.insert:
			#print 'insert moved!', self.tag_widgets
			for tags,widg in self.tag_widgets.items():
				active = True
				for t in tags:
					#print 'testing tag',t
					if not iter.has_tag(t):
						#print "iter doesn't have tag %s!"%t
						active=False
				self.internal_toggle=True
				widg.set_active(active)
				self.internal_toggle=False


	def _value_get(self):
		return self.get_text()

	def _value_set(self, value):
		if value==False:
			value=''
		buffer = self.tv.get_buffer()
		buffer.delete(buffer.get_start_iter(), buffer.get_end_iter())
		self.set_text(value)
		return True

	value = property(_value_get, _value_set, None,
	  'The content of the widget or ValueError if not valid')

	def clear(self):
		self.value = ''

==> tinyerp-3.0.1.tar.gz: 3.0.1/client/bin/widget/form/textbox_tag.py <==
##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#					Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

#
# TODO: recode the whole pango code
# 

import gtk
from gtk import glade
import pango

import wid_int
import common

class textbox_tag(wid_int.wid_int):
	desc_to_attr_table = {
		'family':[pango.AttrFamily,""],
		'style':[pango.AttrStyle,pango.STYLE_NORMAL],
		'variant':[pango.AttrVariant,pango.VARIANT_NORMAL],
		'weight':[pango.AttrWeight,pango.WEIGHT_NORMAL],
		'stretch':[pango.AttrStretch,pango.STRETCH_NORMAL],
		}
	pango_translation_properties={
			pango.ATTR_SIZE : 'size',
			pango.ATTR_WEIGHT: 'weight',
			pango.ATTR_UNDERLINE: 'underline',
			pango.ATTR_STRETCH: 'stretch',
			pango.ATTR_VARIANT: 'variant',
			pango.ATTR_STYLE: 'style',
			pango.ATTR_SCALE: 'scale',
			pango.ATTR_STRIKETHROUGH: 'strikethrough',
			pango.ATTR_RISE: 'rise',
			}
	attval_to_markup={
			'underline':{pango.UNDERLINE_SINGLE:'single',
						 pango.UNDERLINE_DOUBLE:'double',
						 pango.UNDERLINE_LOW:'low',
						 pango.UNDERLINE_NONE:'none'},
			'stretch':{pango.STRETCH_ULTRA_EXPANDED:'ultraexpanded',
					   pango.STRETCH_EXPANDED:'expanded',
					   pango.STRETCH_EXTRA_EXPANDED:'extraexpanded',
					   pango.STRETCH_EXTRA_CONDENSED:'extracondensed',
					   pango.STRETCH_ULTRA_CONDENSED:'ultracondensed',
					   pango.STRETCH_CONDENSED:'condensed',
					   pango.STRETCH_NORMAL:'normal',
					   },
			'variant':{pango.VARIANT_NORMAL:'normal',
					   pango.VARIANT_SMALL_CAPS:'smallcaps',
					   },
			'style':{pango.STYLE_NORMAL:'normal',
					 pango.STYLE_OBLIQUE:'oblique',
					 pango.STYLE_ITALIC:'italic',
					 },
			'stikethrough':{1:'true',
							True:'true',
							0:'false',
							False:'false'},
			}

	def __init__(self, parent, attrs={}):
		wid_int.wid_int.__init__(self, parent, attrs)
		self.win_gl = glade.XML(common.terp_path("terp.glade"),"widget_textbox_tag")
		self.widget = self.win_gl.get_widget('widget_textbox_tag')
		self.tv = self.win_gl.get_widget('widget_textbox_tag_tv')
		self.widget.show_all()
		self.tags = {}
		self.buf = self.tv.get_buffer()
		self.tagdict = {}


		#a,t,s = pango.parse_markup('<b>bold</b>',u'0')

		#ctx = self.widget.create_pango_context()
		#font = pango.FontDescription("sans bold 12")
		
		#tags = self.get_tags_from_attrs(font,None,[])
		#self.tag_widgets[tuple(tags)]=widg
		self.internal_toggle = False
		self.setup_default_tags ()
		self.win_gl.signal_connect('on_toggle_bold_toggled', self._toggle, [self.bold])
		self.win_gl.signal_connect('on_toggle_italic_toggled', self._toggle, [self.italic])
		self.win_gl.signal_connect('on_toggle_underline_toggled', self._toggle, [self.underline])


	def get_tags_from_attrs (self, font,lang,attrs):
		tags = []
		if font:
			font,fontattrs = self.fontdesc_to_attrs(font)
			fontdesc = font.to_string()
			if fontattrs:
				attrs.extend(fontattrs)
			if fontdesc and fontdesc!='Normal':
				if not self.tags.has_key(font.to_string()):
					tag=self.buf.create_tag()
					tag.set_property('font-desc',font)
					if not self.tagdict.has_key(tag): self.tagdict[tag]={}
					self.tagdict[tag]['font_desc']=font.to_string()
					self.tags[font.to_string()]=tag
				tags.append(self.tags[font.to_string()])
		if lang:
			if not self.tags.has_key(lang):
				tag = self.buf.create_tag()
				tag.set_property('language',lang)
				self.tags[lang]=tag
			tags.append(self.tags[lang])
		if attrs:
			for a in attrs:
				if a.type == pango.ATTR_FOREGROUND:
					gdkcolor = self.pango_color_to_gdk(a.color)
					key = 'foreground%s'%self.color_to_hex(gdkcolor)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property('foreground-gdk',gdkcolor)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]]['foreground']="#%s"%self.color_to_hex(gdkcolor)
					tags.append(self.tags[key])
				if a.type == pango.ATTR_BACKGROUND:
					gdkcolor = self.pango_color_to_gdk(a.color)
					tag.set_property('background-gdk',gdkcolor)
					key = 'background%s'%self.color_to_hex(gdkcolor)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property('background-gdk',gdkcolor)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]]['background']="#%s"%self.color_to_hex(gdkcolor)
					tags.append(self.tags[key])
				if self.pango_translation_properties.has_key(a.type):
					prop=self.pango_translation_properties[a.type]
					#print 'setting property %s of %s (type: %s)'%(prop,a,a.type)
					val=getattr(a,'value')
					#tag.set_property(prop,val)
					mval = val
					if self.attval_to_markup.has_key(prop):
						#print 'converting ',prop,' in ',val
						if self.attval_to_markup[prop].has_key(val):
							mval = self.attval_to_markup[prop][val]
					key="%s%s"%(prop,val)
					if not self.tags.has_key(key):
						self.tags[key]=self.buf.create_tag()
						self.tags[key].set_property(prop,val)
						self.tagdict[self.tags[key]]={}
						self.tagdict[self.tags[key]][prop]=mval
					tags.append(self.tags[key])
				else:
					print "Don't know what to do with attr %s"%a
		return tags

	def get_tags (self):
		tagdict = {}
		for pos in range(self.buf.get_char_count()):
			iter=self.buf.get_iter_at_offset(pos)
			for tag in iter.get_tags():
				if tagdict.has_key(tag):
					if tagdict[tag][-1][1] == pos - 1:
						tagdict[tag][-1] = (tagdict[tag][-1][0],pos)
					else:
						tagdict[tag].append((pos,pos))
				else:
					tagdict[tag]=[(pos,pos)]
		return tagdict

	def get_text (self):
		tagdict=self.get_tags()
		txt = self.buf.get_text(self.buf.get_start_iter(),self.buf.get_end_iter())
		cuts = {}
		for k,v in tagdict.items():
			stag,etag = self.tag_to_markup(k)
			for st,end in v:
				if cuts.has_key(st): cuts[st].append(stag) #add start tags second
				else: cuts[st]=[stag]
				if cuts.has_key(end+1): cuts[end+1]=[etag]+cuts[end+1] #add end tags first
				else: cuts[end+1]=[etag]
		last_pos = 0
		outbuff = ""
		cut_indices = cuts.keys()
		cut_indices.sort()
		for c in cut_indices:
			if not last_pos==c:
				outbuff += txt[last_pos:c]
				last_pos = c
			for tag in cuts[c]:
				outbuff += tag
		outbuff += txt[last_pos:]
		return outbuff

	def tag_to_markup (self, tag):
		stag = "<span"
		for k,v in self.tagdict[tag].items():
			stag += ' %s="%s"'%(k,v)
		stag += ">"
		return stag,"</span>"

	def fontdesc_to_attrs (self,font):
		nicks = font.get_set_fields().value_nicks
		attrs = []
		for n in nicks:
			if self.desc_to_attr_table.has_key(n):
				Attr,norm = self.desc_to_attr_table[n]
				# create an attribute with our current value
				attrs.append(Attr(getattr(font,'get_%s'%n)()))
				# unset our font's value
				getattr(font,'set_%s'%n)(norm)
		return font,attrs

	def pango_color_to_gdk (self, pc):
		return gtk.gdk.Color(pc.red,pc.green,pc.blue)

	def color_to_hex (self, color):
		hexstring = ""
		for col in 'red','green','blue':
			hexfrag = hex(getattr(color,col)/(16*16)).split("x")[1]
			if len(hexfrag)<2: hexfrag = "0" + hexfrag
			hexstring += hexfrag
		return hexstring

	def apply_font_and_attrs (self, font, attrs):
		tags = self.get_tags_from_attrs(font,None,attrs)
		for t in tags: self.apply_tag(t)

	def remove_font_and_attrs (self, font, attrs):
		tags = self.get_tags_from_attrs(font,None,attrs)
		for t in tags: self.remove_tag(t)

	def setup_default_tags (self):
		self.italic = self.get_tags_from_attrs(None,None,[pango.AttrStyle('italic')])[0]
		self.bold = self.get_tags_from_attrs(None,None,[pango.AttrWeight('bold')])[0]
		self.underline = self.get_tags_from_attrs(None,None,[pango.AttrUnderline('single')])[0]

	def get_selection (self):
		bounds = self.buf.get_selection_bounds()
		if not bounds:
			iter=self.buf.get_iter_at_mark(self.buf.get_mark('insert'))
			if iter.inside_word():
				start_pos = iter.get_offset()
				iter.forward_word_end()
				word_end = iter.get_offset()
				iter.backward_word_start()
				word_start = iter.get_offset()
				iter.set_offset(start_pos)
				bounds = (self.buf.get_iter_at_offset(word_start),
						  self.buf.get_iter_at_offset(word_end))
		return bounds

	def apply_tag (self, tag):
		selection = self.get_selection()
		if selection:
			self.buf.apply_tag(tag,*selection)
		else:
			insert_iter = self.buf.get_iter_at_mark(self.buf.get_mark('insert'))
			self.buf.apply_tag(tag,insert_iter, insert_iter)
		
	def remove_tag (self, tag):
		selection = self.get_selection()
		if selection:
			self.buf.remove_tag(tag,*selection)

	def remove_all_tags (self):
		selection = self.get_selection()
		if selection:
			for t in self.tags.values():
				self.buf.remove_tag(t,*selection)

	def set_text (self, txt):
		buf = self.tv.get_buffer()
		parsed,txt,separator = pango.parse_markup(txt,u'0')
		attrIter = parsed.get_iterator()
		while True:
			range=attrIter.range()
			font,lang,attrs = attrIter.get_font()
			tags = self.get_tags_from_attrs(font,lang,attrs)
			text = txt[range[0]:range[1]]
			if tags:
				buf.insert_with_tags(buf.get_end_iter(),text,*tags)
			else:
				buf.insert_with_tags(buf.get_end_iter(),text)
			if not attrIter.next():
				break

	#
	# Button callback
	#
	def setup_widget_from_pango (self, widg, markupstring):
		"""setup widget from a pango markup string"""
		#font = pango.FontDescription(fontstring)
		a,t,s = pango.parse_markup(markupstring,u'0')
		ai=a.get_iterator()
		font,lang,attrs=ai.get_font()
		self.setup_widget(widg,font,attrs)

	def setup_widget (self, widg, font, attr):
		tags=self.get_tags_from_attrs(font,None,attr)
		self.tag_widgets[tuple(tags)]=widg
		widg.connect('toggled',self._toggle,tags)

	def _toggle (self, widget, tags):
		if self.internal_toggle: return
		#print 'not internal'
		if widget.get_active():
			#print 'applying'
			for t in tags: self.apply_tag(t)
		else:
			#print 'removing'
			for t in tags: self.remove_tag(t)

	def _mark_set_cb (self, buffer, iter, mark, *params):
		#print '_mark_set_cb'
		if mark==self.insert:
			#print 'insert moved!', self.tag_widgets
			for tags,widg in self.tag_widgets.items():
				active = True
				for t in tags:
					#print 'testing tag',t
					if not iter.has_tag(t):
						#print "iter doesn't have tag %s!"%t
						active=False
				self.internal_toggle=True
				widg.set_active(active)
				self.internal_toggle=False


	def _value_get(self):
		return self.get_text()

	def _value_set(self, value):
		if value==False:
			value=''
		buffer = self.tv.get_buffer()
		buffer.delete(buffer.get_start_iter(), buffer.get_end_iter())
		self.set_text(value)
		return True

	value = property(_value_get, _value_set, None,
	  'The content of the widget or ValueError if not valid')

	def clear(self):
		self.value = ''
